import pytest
from datetime import datetime
import pandas as pd


from ..catalog import Catalog, get_file_relative_path


@pytest.fixture
def catalog2():
    return Catalog(release_tag="2.0")


@pytest.fixture
def catalog_latest():
    return Catalog(release_tag="latest")


@pytest.fixture
def catalog_empty():
    return Catalog()


@pytest.fixture
def catalog_df():
    df = pd.DataFrame(
        {
            "NAXIS1": [12, 15, 20],
            "NAXIS2": [100, 101, 102],
            "NAXIS3": [41, 44, 47],
            "NAXIS4": [1, 1, 1],
            "OBT_BEG": [0, 0, 0],
            "LEVEL": ["L2", 0, 0],
            "FILENAME": [0, 0, 0],
            "DATE-BEG": [pd.Timestamp("2023-02-01T12:34"), 0, 0],
            "SPIOBSID": [0, 0, 0],
            "RASTERNO": [0, 0, 0],
            "STUDYTYP": [0, 0, 0],
            "MISOSTUD": [0, 0, 0],
            "XPOSURE": [0, 0, 0],
            "CRVAL1": [0, 0, 0],
            "CDELT1": [0, 0, 0],
            "CRVAL2": [0, 0, 0],
            "CDELT2": [0, 0, 0],
            "STP": [0, 0, 0],
            "DSUN_AU": [0, 0, 0],
            "CROTA": [0, 0, 0],
            "OBS_ID": [0, 0, 0],
            "SOOPNAME": [0, 0, 0],
            "SOOPTYPE": [0, 0, 0],
            "NWIN": [0, 0, 0],
            "DARKMAP": [0, 0, 0],
            "COMPLETE": [0, 0, 0],
            "SLIT_WID": [0, 0, 0],
            "DATE": [0, 0, 0],
            "PARENT": [0, 0, 0],
            "HGLT_OBS": [0, 0, 0],
            "HGLN_OBS": [0, 0, 0],
            "PRSTEP1": [0, 0, 0],
            "PRPROC1": [0, 0, 0],
            "PRPVER1": [0, 0, 0],
            "PRPARA1": [0, 0, 0],
            "proc_steps": [
                '[{"PRPROC": "JPEG Compression (On-board)"},'
                ' {"PRPROC": "spice_prep_dark_offset_correction.pro",'
                ' "PRPVER": "1.3", "PRPARA": "dark_spiobsid=117440828"}]',
                0,
                0,
            ],
        }
    )
    return Catalog(data_frame=df)


def test_get_file_relative_path(catalog_df):
    path = get_file_relative_path(catalog_df.iloc[0])
    assert path.as_posix() == "level2/2023/02/01"


class TestCatalog:
    def test_init_filename(self):
        filename = "wepocmwkx.fts"
        with pytest.raises(RuntimeError, match="does not exist"):
            catalog = Catalog(filename)  # noqa: F841

    def test_init_release(self, catalog2, catalog_latest):
        columns = {
            "NAXIS1",
            "NAXIS2",
            "NAXIS3",
            "NAXIS4",
            "OBT_BEG",
            "LEVEL",
        }
        assert columns.issubset(catalog2.columns)
        assert len(catalog2) > 10000
        assert columns.issubset(catalog_latest.columns)
        assert len(catalog_latest) > 10000
        assert catalog_latest._cache_release_catalog() is None

    def test_init_empty(self, catalog_empty):
        assert len(catalog_empty) == 0
        assert len(catalog_empty.columns) == 0

    def test_init_dataframe(self, catalog_df):
        assert len(catalog_df) == 3
        assert len(catalog_df.columns) == 36
        assert catalog_df.iloc[1].NAXIS2 == 101

    def test_find_file(self, catalog2):
        expected_filename = "solo_L2_spice-n-exp_20211204T120022_V02_83886365-000.fits"
        result = catalog2.find_file(pd.Timestamp("2021-10-10"))
        assert result.FILENAME == expected_filename
        result = catalog2.find_file(datetime(2021, 10, 10))
        assert result.FILENAME == expected_filename
        result = catalog2.find_file("2021-10-10")
        assert result.FILENAME == expected_filename
        result = catalog2.find_file(None)
        assert result is None



import pytest

from ..release import Release, get_latest_release_tag


@pytest.fixture
def release2():
    return Release("2.0")


class TestRelease:
    def test_init(self, release2):
        assert release2.tag == "2.0"
        assert release2.base_url.startswith(
            "https://spice.osups.universite-paris-saclay.fr"
        )
        assert release2.url.endswith("2.0/")
        assert release2.catalog_url.endswith("/catalog.csv")
        assert release2.exists  # online access
        assert not release2.is_latest  # online access

    def test_latest(self):  # online access
        release = Release()
        assert release.tag != "2.0"
        assert release.base_url.startswith(
            "https://spice.osups.universite-paris-saclay.fr"
        )
        assert release.catalog_url.endswith("/catalog.csv")
        assert release.exists
        assert release.is_latest

    def test_get_latest(self):
        assert get_latest_release_tag() != "2.0"



import numpy as np


def rss(a, axis=None):
    """
    Root sum square of array elements

    Parameters
    ----------
    a: numpy.array
        Input values
    axis: None or int or tuple of ints
        Axis or axes along which the root-sum-square is performed.
    Return
    ------
    float
        Root sum square of input values over a given axis
    """
    return np.linalg.norm(a, axis=axis)



from dataclasses import dataclass
import requests


@dataclass
class Release:
    tag: str = None
    base_url: str = "https://spice.osups.universite-paris-saclay.fr/spice-data/"
    _latest_tag: str = None

    def __post_init__(self):
        """
        Initialize object to latest release if no release tag set
        """
        if self.tag is None:
            self.tag = self.latest_tag

    @property
    def url(self):
        """
        Return
        ------
        str
            Release URL
        """
        return f"{self.base_url}release-{self.tag}/"

    @property
    def catalog_url(self):
        """
        Return
        ------
        str
            Catalog URL for release
        """
        return self.url + "catalog.csv"

    @property
    def latest_tag(self):
        """
        Return
        ------
        str
            Tag of latest release
        """
        if self._latest_tag is None:
            url = f"{self.base_url}metadata/latest-release.txt"
            result = requests.get(url)
            if not result.ok:
                raise RuntimeError(
                    "Could not access URL for file with latest release tag"
                )
            self._latest_tag = result.text.split("\n")[0]
        return self._latest_tag

    @property
    def is_latest(self):
        """
        Return
        ------
        bool
            True is object corresponds to latest release
        """
        return self.tag == self.latest_tag

    @property
    def exists(self):
        """
        Return
        ------
        bool
            True if release exists (is accessible online)
        """
        result = requests.get(self.url)
        return result.ok


def get_latest_release_tag():
    """
    Return
    ------
    str
        Latest available release tag
    """
    release = Release()
    return release.latest_tag



import pytest

import numpy as np
import astropy.units as u

from ..spice import Spice


@pytest.fixture
def spice():
    return Spice()


class TestSpice:
    def test_init(self, spice):
        assert spice.read_noise == 6.9 * u.ct / u.pix
        assert spice.aeff_data is None

    def test_effective_area(self, spice):
        # Expected values in mm² for wavelengths in nm
        expected = {
            20: np.nan,
            77: 4.33579493,  # Ne VIII (SW)
            102.6: 9.57706423,  # H Lyβ (LW)
            52.1: 0.43514445,  # Si XII (LW, 2nd order)
            200: np.nan,
        }
        for wavelength in expected:
            assert u.isclose(
                spice.effective_area(wavelength * u.nm),
                expected[wavelength] * u.mm**2,
                equal_nan=True,
            )
        assert u.allclose(
            spice.effective_area([50, 80, 100] * u.nm),
            [0.28686351, 5.50425673, 9.21953583] * u.mm**2,
        )
        assert spice.aeff_data is not None

    def test_quantum_efficiency(self, spice):
        # Expected values for wavelengths in nm
        # Not sure about the value for 2nd order LW
        expected = {
            77: 0.1,
            102.6: 0.25,
        }
        for wavelength in expected:
            assert u.isclose(
                spice.quantum_efficiency(wavelength * u.nm), expected[wavelength]
            )

    def test_which_detector(self, spice):
        # Expected values for wavelengths in nm
        expected = {
            77: "SW",
            102.6: "LW",
        }
        for wavelength in expected:
            assert spice.which_detector(wavelength * u.nm) == expected[wavelength]
        # Expected None values
        expected_none = [70, 85, 106]
        for wavelength in expected_none:
            assert spice.which_detector(wavelength * u.nm) is None

    def test_gain(self, spice):
        # Expected values in DN/ph for wavelengths in nm
        expected = {
            20: np.nan,
            77: 3.58,
            102.6: 0.57,
            200: np.nan,
        }
        for wavelength in expected:
            assert u.isclose(
                spice.gain(wavelength * u.nm),
                expected[wavelength] * u.ct / u.ph,
                equal_nan=True,
            )

    def test_dark_current(self, spice):
        # Expected values in DN/s/pix for wavelengths in nm
        expected = {
            20: np.nan,
            77: 0.89,
            102.6: 0.54,
            200: np.nan,
        }
        for wavelength in expected:
            assert u.isclose(
                spice.dark_current(wavelength * u.nm),
                expected[wavelength] * u.ct / u.s / u.pix,
                equal_nan=True,
            )

    def test_noise_factor(self, spice):
        # Expected values in DN/s/pix for wavelengths in nm
        expected = {
            20: np.nan,
            77: 1.0,
            102.6: 1.6,
            200: np.nan,
        }
        for wavelength in expected:
            assert u.isclose(
                spice.noise_factor(wavelength * u.nm),
                expected[wavelength],
                equal_nan=True,
            )



import astropy.units as u

from ..instrument_modelling import Spice, Study, Observation


def spice_error(hdu=None, data=None, header=None, verbose=True):
    """
    Return total (measured) signal increase and standard deviation due to noises

    Parameters
    ----------
    hdu: astropy.io.fits.hdu.image.ImageHDU
        SPICE L2 FITS HDU
    data: numpy.ndarray
        SPICE L2 FITS data, assumed to be in W / m2 / sr / nm
    header: astropy.io.fits.header.Header
        SPICE L2 FITS header
    verbose: bool
        If True, displays details

    Return
    ------
    float:
        Average contribution of noise to measured signal
    dict:
        Noise standard deviations for the different components (and total)

    Either hdu, or data and header should be provided.
    """
    if data is None or header is None:
        if hdu is None:
            raise RuntimeError("Either hdu, or data and header should be provided")
        header = hdu.header
        data = hdu.data
    if header["LEVEL"] != "L2":
        raise RuntimeError("Level should be L2")
    data *= u.Unit(header["BUNIT"])
    print(data.unit)
    study = Study()
    study.init_from_header(header)
    if verbose:
        print(f"Getting observation parameters from {header['EXTNAME']}")
        print(study)
    instrument = Spice()
    observation = Observation(instrument, study)
    av_constant_noise_level, sigma = observation.noise_effects_from_l2(
        data, study.av_wavelength
    )
    return av_constant_noise_level, sigma



import pytest

from astropy.io import fits
import astropy.units as u

from ..uncertainties import spice_error


specrad_unit = u.mW / u.m**2 / u.sr / u.nm


@pytest.fixture
def hdus():
    # all expected values in specrad_unit
    hdus = [
        {
            "url": "https://spice.osups.universite-paris-saclay.fr/spice-data/release-3.0/level2/2022/04/02/solo_L2_spice-n-ras_20220402T111537_V06_100664002-000.fits",
            "hdu_index": 2,  # Ne VIII window
            "expected_constant": 42.9513921,
            "pixel_index": (0, 25, 415, 80),
            "expected": {
                "Dark": 20.36091256,
                "Background": 0.0,
                "Read": 66.59878776,
                "Signal": 168.60160232,
                "Total": 182.41837621,
            },
        },
        {  # File used for comparison with MPS' IDL code
            "url": "https://spice.osups.universite-paris-saclay.fr/spice-data/release-3.0/level2/2021/02/23/solo_L2_spice-n-ras_20210223T154400_V10_50331754-000.fits",
            "hdu_index": 2,  # Lyγ - C III
            "expected_constant": 87.80646,
            "pixel_index": (0, 25, 415, 80),
            "expected": {
                "Dark": 26.718655,
                "Background": 0.0,
                "Read": 56.098571,
                "Signal": 61.990004,
                "Total": 87.7707,
            },
        },
    ]
    for item in hdus:
        item["hdu_list"] = fits.open(item["url"])
    yield hdus
    for item in hdus:
        item["hdu_list"].close()


class TestUncertainties:
    def test_spice_error(self, hdus):
        for item in hdus:
            av_constant_noise_level, sigma = spice_error(
                item["hdu_list"][item["hdu_index"]]
            )
        assert u.isclose(
            av_constant_noise_level, item["expected_constant"] * specrad_unit
        )
        for component in item["expected"]:
            pixel_index = (
                () if len(sigma[component].shape) == 0 else item["pixel_index"]
            )
            print(f"{sigma[component][pixel_index]=}")
            assert u.isclose(
                sigma[component][pixel_index],
                item["expected"][component] * specrad_unit,
            )



from dataclasses import dataclass
import astropy.units as u


@dataclass
class Study:
    """
    Study parameters
    """

    slit: u.arcsec = None
    bin_x: int = None  # bin over x or wavelength axis
    bin_y: int = None
    window_width: u.pix = None
    exp_time: u.s = None
    av_wavelength: u.m = None
    radcal: u.ct / (u.W / u.m**2 / u.sr / u.nm) = None
    level: str = None

    def init_from_header(self, header):
        """
        Initialize study parameters from FITS header

        Parameters
        ----------
        header: astropy.io.fits.Header
            FITS header
        """
        # TODO use real slit width, not nominal slit width
        self.slit = header["SLIT_WID"] * u.arcsec
        self.bin_x = header["NBIN3"]  # bin factor in dispersion direction
        self.bin_y = header["NBIN2"]  # bin factor in slit direction
        self.exp_time = header["XPOSURE"] * u.s
        self.window_width = header["NAXIS3"] * u.pix
        self.av_wavelength = (
            (header["WAVEMIN"] + header["WAVEMAX"]) / 2 * 10 ** header["WAVEUNIT"] * u.m
        )
        self.level = header["LEVEL"]
        if self.level == "L2":
            self.radcal = header["RADCAL"] * u.ct / (u.W / u.m**2 / u.sr / u.nm)
        else:
            self.radcal = None  # TODO or need to have a value of 1?

    def __str__(self):
        if self.slit is None:
            return "Non-initialized study"
        else:
            return f"""
Slit: {self.slit}
Bin: ({self.bin_x}, {self.bin_y})
Exposure time: {self.exp_time}
Window width: {self.window_width}
Average wavelength: {self.av_wavelength.to(u.nm)}
RADCAL: {self.radcal}
            """



import pytest

from astropy.io import fits
import astropy.units as u

from ..spice import Spice
from ..study import Study
from ..observation import Observation


@pytest.fixture
def header():
    return {
        "SLIT_WID": 4.0,
        "NBIN3": 1,
        "NBIN2": 2,
        "XPOSURE": 10.0,
        "NAXIS3": 48,
        "WAVEMIN": 769,
        "WAVEMAX": 771,
        "WAVEUNIT": -10,
        "LEVEL": "L2",
        "RADCAL": 1000.0,
    }


@pytest.fixture
def hdu():
    url = "https://spice.osups.universite-paris-saclay.fr/spice-data/release-3.0/level2/2022/04/02/solo_L2_spice-n-ras_20220402T111537_V06_100664002-000.fits"
    # with fits.open(url) as hdu_list:
    hdu_list = fits.open(url)
    hdu = hdu_list[2]  # Ne VIII window
    yield hdu
    hdu_list.close()


@pytest.fixture
def observation(header):
    instrument = Spice()
    study = Study()
    study.init_from_header(header)
    return Observation(instrument, study)


class TestObservation:
    def test_observation_from_spice_hdu(self, hdu):
        observation = Observation.observation_from_spice_hdu(hdu)
        assert type(observation.instrument) is Spice
        assert type(observation.study) is Study

    def test_av_dark_current(self, observation):
        # Expected values in DN/ph for wavelengths in nm
        expected = {
            77: 17.8,
            102.5: 10.8,
        }
        for wavelength in expected:
            assert u.isclose(
                observation.av_dark_current(wavelength * u.nm),
                expected[wavelength] * u.ct / u.pix,
            )

    def test_av_background(self, observation):
        # Expected values in DN/ph for wavelengths in nm
        expected = {
            77: 0,
            102.5: 0,
        }
        for wavelength in expected:
            assert u.isclose(
                observation.av_background(wavelength * u.nm),
                expected[wavelength] * u.ct / u.pix,
            )

    def test_read_noise_width(self, observation):
        assert u.isclose(observation.read_noise_width, 9.75807358 * u.ct / u.pix)

    def test_noise_effects_from_l2(self, observation):
        specrad_unit = u.mW / u.m**2 / u.sr / u.nm
        av_constant_noise_level, sigma = observation.noise_effects_from_l2(
            0.1 * specrad_unit, 77 * u.nm
        )
        assert u.isclose(av_constant_noise_level, 17.8 * specrad_unit)
        expected = {
            "Dark": 4.219004621945797,
            "Background": 0.0,
            "Read": 9.758073580374356,
            "Signal": 18.920887928424502,
            "Total": 21.702995184996933,
        }
        for component in expected:
            assert u.isclose(sigma[component], expected[component] * specrad_unit)



import pytest

import astropy.units as u

from ..study import Study


@pytest.fixture
def empty_study():
    return Study()


@pytest.fixture
def header():
    return {
        "SLIT_WID": 4.0,
        "NBIN3": 1,
        "NBIN2": 2,
        "XPOSURE": 10.0,
        "NAXIS3": 48,
        "WAVEMIN": 769,
        "WAVEMAX": 771,
        "WAVEUNIT": -10,
        "LEVEL": "L2",
        "RADCAL": 1000.0,
    }


@pytest.fixture
def study(header):
    s = Study()
    s.init_from_header(header)
    return s


class TestStudy:
    def test_init(self, empty_study):
        attributes = [
            "slit",
            "bin_x",
            "bin_y",
            "window_width",
            "exp_time",
            "av_wavelength",
            "radcal",
        ]
        for attribute in attributes:
            assert getattr(empty_study, attribute) is None

    def test_init_from_header(self, study):
        expected_values = {
            "slit": 4 * u.arcsec,
            "bin_x": 1,
            "bin_y": 2,
            "window_width": 48 * u.pix,
            "exp_time": 10 * u.s,
            "av_wavelength": 77 * u.nm,
            "radcal": 1000 * u.ct * u.m**2 * u.nm * u.sr / u.W,
        }
        for attribute in expected_values:
            assert u.isclose(getattr(study, attribute), expected_values[attribute])

    def test___str__(self, empty_study, study):
        assert str(empty_study) == "Non-initialized study"
        assert type(str(study)) is str



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
from datetime import datetime

# from sospice import __version__
from setuptools_scm import get_version


# -- Project information -----------------------------------------------------

project = "sospice"
copyright = f"{datetime.now().year}, SPICE consortium"
author = "SPICE consortium"

# The full version, including alpha/beta/rc tags
release = get_version(root="../..", relative_to=__file__)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "matplotlib.sphinxext.plot_directive",
    "numpydoc",
    # "sphinx_automodapi.automodapi",
    # "sphinx_automodapi.smart_resolver",
    # "sphinx_changelog",
    # "sphinx_gallery.gen_gallery",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    # "sphinxext.opengraph",
    # "sphinx_design",
    # "sphinx_copybutton",
    # "hoverxref.extension",
]

autosummary_generate = True  # Turn on sphinx.ext.autosummary

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["build", "htmlcov", "sospice.egg-info", "venv", "tests"]

# -- Options for intersphinx extension ---------------------------------------
intersphinx_mapping = {
    "python": (
        "https://docs.python.org/3/",
        (None, "http://data.astropy.org/intersphinx/python3.inv"),
    ),
    "numpy": (
        "https://docs.scipy.org/doc/numpy/",
        (None, "http://data.astropy.org/intersphinx/numpy.inv"),
    ),
    "matplotlib": (
        "https://matplotlib.org/",
        (None, "http://data.astropy.org/intersphinx/matplotlib.inv"),
    ),
    "astropy": ("http://docs.astropy.org/en/stable/", None),
    "sunpy": ("https://docs.sunpy.org/en/stable/", None),
    "sunraster": ("https://docs.sunpy.org/projects/sunraster/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



from dataclasses import dataclass
import numpy as np
import astropy.units as u

from .spice import Spice
from .study import Study
from ..util import rss


@dataclass
class Observation:
    instrument: Spice()
    study: Study()

    @classmethod
    def observation_from_spice_hdu(cls, hdu, verbose=True):
        """
        Generate an Observation object from a SPICE L2 file HDU
        """
        study = Study()
        study.init_from_header(hdu.header)
        if verbose:
            print(f"Getting observation parameters from {hdu.name}")
            print(study)
        instrument = Spice()
        observation = Observation(instrument, study)
        return observation

    @u.quantity_input
    def av_dark_current(self, wvl: u.Angstrom = None):
        """
        Average dark current in DN per macro-pixel over exposure time

        Parameters
        ----------
        wvl: Quantity
            Wavelength (or array of wavelengths)

        Return
        ------
        float
            Average dark current

        TODO:
        * Should depend on detector temperature.
        * Actually non-Poissonian (need to look at real darks).
        * Would depend on position (dark pixels) in L1
        """
        if wvl is None:
            wvl = self.study.av_wavelength
        return (
            self.instrument.dark_current(wvl)
            * self.study.exp_time
            * self.study.bin_x
            * self.study.bin_y
        ).to(u.ct / u.pix)

    @u.quantity_input
    def av_background(self, wvl: u.Angstrom = None):
        """
        Average background signal in DN per macro-pixel over exposure time

        Parameters
        ----------
        wvl: Quantity
            Wavelength (or array of wavelengths)

        Return
        ------
        float
            Average background
        """
        if wvl is None:
            wvl = self.study.av_wavelength
        return (
            self.instrument.background
            * self.instrument.quantum_efficiency(wvl)
            * self.study.exp_time
            * self.study.bin_x
            * self.study.bin_y
            * self.instrument.gain(wvl)
        ).to(u.ct / u.pix)

    @property
    def read_noise_width(self):
        """
        Read noise distribution width in DN per macro-pixel
        TODO make sure that this is a standard deviation and not a FWHM
        """
        return self.instrument.read_noise * np.sqrt(self.study.bin_x * self.study.bin_y)

    @u.quantity_input
    def noise_effects(self, signal_mean: u.ct / u.pix, wvl: u.Angstrom = None):
        """
        Return total (measured) signal increase and standard deviation due to noises

        Parameters
        ----------
        signal_mean: Quantity
            Measured signal mean, in DN/pix, excluding expected signal increase
            due to average dark current and background (so this is not exactly
            the measured signal).
        wvl: Quantity
            Wavelength (or array of wavelengths)

        Return
        ------
        float:
            Average contribution of noise to measured signal
        dict:
            Noise standard deviations for the different components (and total
            uncertainty resulting from them)

        Negative values of the signal are considered to be 0 for the purpose of
        computing the noise on the signal. However, the total uncertainty is
        then set to |signal_mean| + RSS (other noises), to ensure that the
        error bars are still compatible with expected fitted functions.
        We suggest users to replace large negative values of the signal
        (e.g. < -3 * RSS(other noises)) by NaNs.

        """
        if wvl is None:
            wvl = self.study.av_wavelength
        av_dark_current = self.av_dark_current()
        av_background = self.av_background()
        av_constant_noise_level = av_dark_current + av_background
        sigma = dict()
        gain = self.instrument.gain(wvl)
        sigma["Dark"] = np.sqrt(av_dark_current.value) * u.ct / u.pix
        sigma["Background"] = np.sqrt(av_background * gain).value * u.ct / u.pix
        sigma["Read"] = self.read_noise_width
        signal_mean_nonneg = np.maximum(signal_mean, 0)
        sigma["Signal"] = np.sqrt(signal_mean_nonneg * gain).value * u.ct / u.pix
        sigma["Signal"] *= self.instrument.noise_factor(wvl)
        constant_noises = rss(
            np.array(
                [sigma["Dark"].value, sigma["Background"].value, sigma["Read"].value]
            )
        )
        sigma["Total"] = (
            rss(
                np.array(
                    [
                        constant_noises * np.ones_like(signal_mean.value),
                        sigma["Signal"].value,
                    ]
                ),
                axis=0,
            )
            * sigma["Signal"].unit
        )
        where_neg = signal_mean < 0
        sigma["Total"][where_neg] = (
            -signal_mean[where_neg] + constant_noises * signal_mean.unit
        )
        return av_constant_noise_level, sigma

    @u.quantity_input
    def noise_effects_from_l2(
        self, data: u.W / u.m**2 / u.sr / u.nm, wvl: u.Angstrom
    ):
        """
        Return total (measured) signal increase and standard deviation due to noises

        Parameters
        ----------
        data: Quantity
            L2 data, in W / m2 / sr / nm
        wvl: Quantity
            Wavelength
        Return
        ------
        float:
            Average contribution of noise to measured signal
        dict:
            Noise standard deviations for the different components (and total)
        """
        data_dn = data * self.study.radcal / u.pix
        av_constant_noise_level, sigma = self.noise_effects(data_dn, wvl)
        av_constant_noise_level /= self.study.radcal / u.pix
        for component in sigma:
            sigma[component] /= self.study.radcal / u.pix
        return av_constant_noise_level, sigma



from dataclasses import dataclass
from pathlib import Path
import numpy as np
from scipy.io import readsav
import astropy.units as u


@dataclass
class Spice:
    # These value the values used in Huang et al. 2023
    # (doi:10.1051/0004-6361/202345988 TBC)
    # which are supposed to be the latest values presented by RAL
    # astropy.units.ct (counts) is used for DNs
    read_noise = 6.9 * u.ct / u.pix
    background = 0.0 * u.ph / u.s / u.pix  # 1.0 in SPICE-RAL-RP-0002
    pix_x = 1.0 * u.arcsec / u.pix  # not used
    pix_y = 1.0 * u.arcsec / u.pix  # not used except for PSF
    pix_w = 0.0095 * u.nm / u.pix  # not used except for PSF
    aeff_data = None  # Will be read from file when needed

    @u.quantity_input
    def effective_area(self, wvl: u.nm):
        """
        Get the SPICE effective area for some wavelength

        Parameters
        ----------
        wvl: Quantity
            Wavelength (or array of wavelengths)

        Return
        ------
        Quantity:
            Effective area(s)
        """
        wvl = wvl.to(u.nm).value
        if self.aeff_data is None:
            code_path = (Path(__file__) / ".." / "..").resolve()
            aeff_file = code_path / "data" / "calibration" / "effective_area.sav"
            self.aeff_data = readsav(aeff_file.as_posix())
        # try interpolating for both second (1) and first (2) order
        left_right = {"left": np.nan, "right": np.nan}
        aeff1 = np.interp(
            wvl, self.aeff_data["lam_1"], self.aeff_data["net_resp_1"], **left_right
        )
        aeff2 = np.interp(
            wvl, self.aeff_data["lam_2"], self.aeff_data["net_resp_2"], **left_right
        )
        # choose where interpolation was done for the correct order
        return np.where(np.isfinite(aeff2), aeff2, aeff1) * u.mm**2

    @u.quantity_input
    def quantum_efficiency(self, wvl: u.Angstrom):
        """
        Get the SPICE detector quantum efficiency for some wavelength

        Parameters
        ----------
        wvl: Quantity
            Wavelength (or array of wavelengths)

        Return
        ------
        Quantity:
            Quantum efficiency(ies): number of detected photons / number of
            incident photons

        Not sure about the values for LW 2nd order, and on what wavelength
        ranges the output should be NaN.
        """
        wvl = wvl.to(u.Angstrom).value
        # Source: Table 8.25 of SPICE-RAL-RP-0002 v10.0
        qe_sw = np.interp(
            wvl,
            [703, 706, 770, 790],  # angstrom
            [0.12, 0.12, 0.1, 0.1],  # electron/photon
            left=np.nan,
            right=np.nan,
        )
        qe_lw = 0.25  # electron/photon
        return np.where((wvl > 703) & (wvl < 791), qe_sw, qe_lw)

    @u.quantity_input
    def which_detector(self, wvl: u.Angstrom):
        """
        Determine which detector corresponds to some wavelength

        Parameters
        ----------
        wvl: Quantity
            Wavelength

        Return
        ------
        str
            Detector name (None if not on a detector)
        """
        wvl = wvl.to(u.Angstrom).value
        if 703 < wvl < 791:
            return "SW"
        elif 970 < wvl < 1053:
            return "LW"
        else:
            return None

    @u.quantity_input
    def gain(self, wvl: u.Angstrom):
        """
        Detector gain as a function of wavelength

        Parameters
        ----------
        wvl: Quantity
            Wavelength

        Return
        ------
        float
            Detector gain
        """
        detector = self.which_detector(wvl)
        if detector is None:
            return np.nan * u.ct / u.ph
        else:
            return {"SW": 3.58, "LW": 0.57}[detector] * u.ct / u.ph

    @u.quantity_input
    def dark_current(self, wvl: u.Angstrom):
        """
        Detector dark current as a function of wavelength

        Parameters
        ----------
        wvl: Quantity
            Wavelength

        Return
        ------
        float
            Detector dark current
        """
        detector = self.which_detector(wvl)
        if detector is None:
            return np.nan * u.ct / u.s / u.pix
        else:
            return {"SW": 0.89, "LW": 0.54}[detector] * u.ct / u.s / u.pix

    @u.quantity_input
    def noise_factor(self, wvl: u.Angstrom):
        """
        Detector noise multiplication factor as a function of wavelength

        Parameters
        ----------
        wvl: Quantity
            Wavelength

        Return
        ------
        float
            Noise multiplication factor
        """
        detector = self.which_detector(wvl)
        if detector is None:
            return np.nan
        else:
            return {"SW": 1.0, "LW": 1.6}[detector]



import numpy as np
from ..rss import rss


def test_rss():
    a = np.array([3, 4])
    assert np.isclose(rss(a), 5.0)
    a = np.array([[3, 4, 5], [5, 4, 3]])
    assert np.isclose(rss(a), 10.0)
    assert np.allclose(rss(a, axis=0), np.sqrt([34, 32, 34]))
    assert np.allclose(rss(a, axis=1), np.sqrt([50, 50]))



from dataclasses import dataclass

import pandas as pd
from pathlib import Path
from astropy.utils.data import download_file

from .release import Release


def get_file_relative_path(cat_row):
    """
    Get file relative path for a given catalog entry

    Parameters
    ----------
    cat_row: pandas.Series
        SPICE catalog entry

    Return
    ------
    pathlib.PosixPath
        File path, relative to the "fits" directory of the file archive:
        leveln/yyyy/mm/dd

    Requires a DATE-BEG header (so does not work with L0 files).
    """
    date = cat_row["DATE-BEG"]
    return (
        Path(f"level{cat_row.LEVEL[1]}")  # noqa: W503
        / f"{date.year}"  # noqa: W503
        / f"{date.month:02}"  # noqa: W503
        / f"{date.day:02}"  # noqa: W503
    )


@dataclass
class Catalog(pd.DataFrame):
    """
    A SPICE catalog, initialized (in that order) either from a filename, a release tag, or a pandas.DataFrame.

    Parameters
    ----------
    filename: str
        A file name (or URL) for the catalog
    release_tag: str
        A release tag. The catalog is fetched online and dowloaded to the astropy cache.
    data_frame: pandas.DataFrame
        A pandas DataFrame to be used as SPICE catalog. Some basic checks are made to ensure
        that is can be used as a SPICE catalog.
    update_cache: bool
        Update cached catalog for the given release tag
    """

    filename: str = None
    release_tag: str = None
    data_frame: str = None
    update_cache: bool = False

    def __post_init__(self):
        """
        Read catalog and update object
        """
        self._normalize_arguments()
        if self.release_tag is not None:
            self._cache_release_catalog()
        if self.filename is not None:
            super().__init__(self.read_catalog())
        else:
            if self.data_frame is None:
                self.data_frame = pd.DataFrame()
            self._validate_data_frame()
            super().__init__(self.data_frame)
            del self.data_frame  # needed for memory usage?
            self.data_frame = None

    def _normalize_arguments(self):
        """
        Prioritize filename then release tag then data frame
        """
        if self.filename is not None:
            self.release_tag = None
            self.data_frame = None
        elif self.release_tag is not None:
            self.data_frame = None

    def _cache_release_catalog(self):
        """
        Used cached catalog or download release catalog to astropy cache
        """
        print("Enter cache release catalog")
        if self.release_tag is None:
            return
        if self.release_tag == "latest":
            self.release_tag = None
        release = Release(self.release_tag)
        assert release.exists
        self.filename = download_file(release.catalog_url, cache=True)
        self.release_tag = None

    def _validate_data_frame(self):
        """
        Check that the data_frame argument can be considered a valid SPICE catalog (or raise an exception)
        """
        print(f"enter validate data frame {self.data_frame}")
        assert self.data_frame is not None
        if self.data_frame.empty:
            return True  # an empty data frame is valid
        required_columns = {
            "NAXIS1",
            "NAXIS2",
            "NAXIS3",
            "NAXIS4",
            "OBT_BEG",
            "LEVEL",
            "FILENAME",
            "DATE-BEG",
            "SPIOBSID",
            "RASTERNO",
            "STUDYTYP",
            "MISOSTUD",
            "XPOSURE",
            "CRVAL1",
            "CDELT1",
            "CRVAL2",
            "CDELT2",
            "STP",
            "DSUN_AU",
            "CROTA",
            "OBS_ID",
            "SOOPNAME",
            "SOOPTYPE",
            "NWIN",
            "DARKMAP",
            "COMPLETE",
            "SLIT_WID",
            "DATE",
            "PARENT",
            "HGLT_OBS",
            "HGLN_OBS",
            "PRSTEP1",
            "PRPROC1",
            "PRPVER1",
            "PRPARA1",
        }
        assert required_columns.issubset(self.data_frame.columns)

    def read_catalog(self):
        """
        Read SPICE FITS files catalog

        Return
        ------
        pandas.DataFrame
            Catalog
        """
        if not Path(self.filename).exists():
            raise RuntimeError(f"File {self.filename} does not exist")
        df = pd.read_csv(
            self.filename,
            low_memory=False,
        )
        date_columns = ["DATE-BEG", "DATE", "TIMAQUTC"]
        for date_column in date_columns:
            df[date_column] = pd.to_datetime(df[date_column], format="ISO8601")
        return df

    def find_file(self, date=None, level="L2"):
        """
        Find closest file to some observation data (DATE-BEG)

        Parameters
        ----------
        date: datetime.datetime, pandas.Timestamp...
            Target date
        level: str
            LEVEL to select (≥L1)

        Return
        ------
        pandas.Series
            File with closest date
        """
        if self.empty or date is None:
            return None
        if type(date) is str:
            date = pd.Timestamp(date)
        df = self[self.LEVEL == level]
        df.set_index("DATE-BEG", inplace=True)
        index = df.index.get_indexer([date], method="nearest")
        df.reset_index(inplace=True)
        return df.iloc[index[0]]
